structure HOL : sig
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype 'a itself = Type
end = struct

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype 'a itself = Type;

end; (*struct HOL*)

structure Fun : sig
  val id : 'a -> 'a
  val fun_upd : 'a HOL.equal -> ('a -> 'b) -> 'a -> 'b -> 'a -> 'b
end = struct

fun id x = (fn xa => xa) x;

fun fun_upd A_ f a b = (fn x => (if HOL.eq A_ x a then b else f x));

end; (*struct Fun*)

structure Arith : sig
  datatype num = One | Bit0 of num | Bit1 of num
  val suc : IntInf.int -> IntInf.int
end = struct

datatype num = One | Bit0 of num | Bit1 of num;

fun suc n = IntInf.+ (n, (1 : IntInf.int));

end; (*struct Arith*)

structure List : sig
  val map : ('a -> 'b) -> 'a list -> 'b list
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  val filter : ('a -> bool) -> 'a list -> 'a list
  val member : 'a HOL.equal -> 'a list -> 'a -> bool
  val insert : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val distinct : 'a HOL.equal -> 'a list -> bool
  val list_all : ('a -> bool) -> 'a list -> bool
  val equal_lista : 'a HOL.equal -> 'a list -> 'a list -> bool
  val equal_list : 'a HOL.equal -> ('a list) HOL.equal
  val removeAll : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val size_list : 'a list -> IntInf.int
end = struct

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun fold f [] = Fun.id
  | fold f (x :: xs) = fold f xs o f x;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun member A_ [] y = false
  | member A_ (x :: xs) y = HOL.eq A_ x y orelse member A_ xs y;

fun insert A_ x xs = (if member A_ xs x then xs else x :: xs);

fun distinct A_ [] = true
  | distinct A_ (x :: xs) = not (member A_ xs x) andalso distinct A_ xs;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    HOL.eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) HOL.equal;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if HOL.eq A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun size_list [] = 0
  | size_list (a :: lista) = IntInf.+ (size_list lista, Arith.suc 0);

end; (*struct List*)

structure Product_Type : sig
  val fst : 'a * 'b -> 'a
  val equal_proda : 'a HOL.equal -> 'b HOL.equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a HOL.equal -> 'b HOL.equal -> ('a * 'b) HOL.equal
  val equal_unita : unit -> unit -> bool
  val equal_unit : unit HOL.equal
end = struct

fun fst (a, b) = a;

fun equal_proda A_ B_ (aa, ba) (a, b) = HOL.eq A_ aa a andalso HOL.eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) HOL.equal;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit HOL.equal;

end; (*struct Product_Type*)

structure Aux : sig
  val distinct_fst : 'a HOL.equal -> ('a * 'b) list -> bool
end = struct

fun distinct_fst A_ = List.distinct A_ o List.map Product_Type.fst;

end; (*struct Aux*)

structure Err : sig
  datatype 'a err = Err | Ok of 'a
end = struct

datatype 'a err = Err | Ok of 'a;

end; (*struct Err*)

structure Option : sig
  val map : ('a -> 'b) -> 'a option -> 'b option
  val the : 'a option -> 'a
  val is_none : 'a option -> bool
  val equal_optiona : 'a HOL.equal -> 'a option -> 'a option -> bool
  val equal_option : 'a HOL.equal -> ('a option) HOL.equal
end = struct

fun map f (SOME x) = SOME (f x)
  | map f NONE = NONE;

fun the (SOME x) = x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun equal_optiona A_ (SOME a) NONE = false
  | equal_optiona A_ NONE (SOME a) = false
  | equal_optiona A_ (SOME aa) (SOME a) = HOL.eq A_ aa a
  | equal_optiona A_ NONE NONE = true;

fun equal_option A_ = {equal = equal_optiona A_} : ('a option) HOL.equal;

end; (*struct Option*)

structure Map : sig
  val map_of : 'a HOL.equal -> ('a * 'b) list -> 'a -> 'b option
  val map_add : ('a -> 'b option) -> ('a -> 'b option) -> 'a -> 'b option
  val map_upds :
    'a HOL.equal -> ('a -> 'b option) -> 'a list -> 'b list -> 'a -> 'b option
end = struct

fun map_of A_ ((l, v) :: ps) k =
  (if HOL.eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun map_add m1 m2 = (fn x => (case m2 x of NONE => m1 x | SOME a => SOME a));

fun map_upds A_ m xs ys = map_add m (map_of A_ (List.rev (List.zip xs ys)));

end; (*struct Map*)

structure Set : sig
  datatype 'a set = Set of 'a list | Coset of 'a list
  val member : 'a HOL.equal -> 'a -> 'a set -> bool
  val remove : 'a HOL.equal -> 'a -> 'a set -> 'a set
  val bot_set : 'a set
  val inf_set : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  val minus_set : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  val union : 'a set -> 'a set -> 'a set
  val insert : 'a -> 'a set -> 'a set
end = struct

datatype 'a set = Set of 'a list | Coset of 'a list;

fun member A_ x (Coset xs) = not (List.member A_ xs x)
  | member A_ x (Set xs) = List.member A_ xs x;

fun remove A_ x (Coset xs) = Coset (List.insert A_ x xs)
  | remove A_ x (Set xs) = Set (List.removeAll A_ x xs);

val bot_set : 'a set = Set [];

fun inf_set A_ a (Coset xs) = List.fold (remove A_) xs a
  | inf_set A_ a (Set xs) = Set (List.filter (fn x => member A_ x a) xs);

fun minus_set A_ a (Coset xs) = Set (List.filter (fn x => member A_ x a) xs)
  | minus_set A_ a (Set xs) = List.fold (remove A_) xs a;

fun union (Set xs) (Set ys) = Set (xs @ ys);

fun insert x (Set xs) = Set (x :: xs);

end; (*struct Set*)

structure Stringa : sig
  datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
    Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC |
    NibbleD | NibbleE | NibbleF
  val equal_literal : string HOL.equal
end = struct

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

val equal_literal = {equal = (fn a => fn b => ((a : string) = b))} :
  string HOL.equal;

end; (*struct Stringa*)

structure Predicate : sig
  datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
  and 'a pred = Seq of (unit -> 'a seq)
  val bind : 'a pred -> ('a -> 'b pred) -> 'b pred
  val apply : ('a -> 'b pred) -> 'a seq -> 'b seq
  val eval : 'a HOL.equal -> 'a pred -> 'a -> bool
  val member : 'a HOL.equal -> 'a seq -> 'a -> bool
  val holds : unit pred -> bool
  val bot_pred : 'a pred
  val single : 'a -> 'a pred
  val adjunct : 'a pred -> 'a seq -> 'a seq
  val sup_pred : 'a pred -> 'a pred -> 'a pred
  val if_pred : bool -> unit pred
  val not_pred : unit pred -> unit pred
end = struct

datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
and 'a pred = Seq of (unit -> 'a seq);

fun bind (Seq g) f = Seq (fn _ => apply f (g ()))
and apply f Empty = Empty
  | apply f (Insert (x, p)) = Join (f x, Join (bind p f, Empty))
  | apply f (Join (p, xq)) = Join (bind p f, apply f xq);

fun eval A_ (Seq f) = member A_ (f ())
and member A_ Empty x = false
  | member A_ (Insert (y, p)) x = HOL.eq A_ x y orelse eval A_ p x
  | member A_ (Join (p, xq)) x = eval A_ p x orelse member A_ xq x;

fun holds p = eval Product_Type.equal_unit p ();

val bot_pred : 'a pred = Seq (fn _ => Empty);

fun single x = Seq (fn _ => Insert (x, bot_pred));

fun adjunct p Empty = Join (p, Empty)
  | adjunct p (Insert (x, q)) = Insert (x, sup_pred q p)
  | adjunct p (Join (q, xq)) = Join (q, adjunct p xq)
and sup_pred (Seq f) (Seq g) =
  Seq (fn _ =>
        (case f () of Empty => g ()
          | Insert (x, p) => Insert (x, sup_pred p (Seq g))
          | Join (p, xq) => adjunct (Seq g) (Join (p, xq))));

fun if_pred b = (if b then single () else bot_pred);

fun not_pred p =
  (if eval Product_Type.equal_unit p () then bot_pred else single ());

end; (*struct Predicate*)

structure Type : sig
  datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty
  val object : string
  val thread : string
  val equal_tya : ty -> ty -> bool
  val equal_ty : ty HOL.equal
  val is_refT_i : ty -> unit Predicate.pred
  val is_refT : ty -> bool
  val throwable : string
  val ground_type : ty -> ty
  val class_type_of : ty -> string option
end = struct

datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty;

val object : string = "java/lang/Object";

val thread : string = "java/lang/Thread";

fun equal_tya (Array ty) (Class literal) = false
  | equal_tya (Class literal) (Array ty) = false
  | equal_tya (Array ty) Nt = false
  | equal_tya Nt (Array ty) = false
  | equal_tya (Class literal) Nt = false
  | equal_tya Nt (Class literal) = false
  | equal_tya (Array ty) Integer = false
  | equal_tya Integer (Array ty) = false
  | equal_tya (Class literal) Integer = false
  | equal_tya Integer (Class literal) = false
  | equal_tya Nt Integer = false
  | equal_tya Integer Nt = false
  | equal_tya (Array ty) Boolean = false
  | equal_tya Boolean (Array ty) = false
  | equal_tya (Class literal) Boolean = false
  | equal_tya Boolean (Class literal) = false
  | equal_tya Nt Boolean = false
  | equal_tya Boolean Nt = false
  | equal_tya Integer Boolean = false
  | equal_tya Boolean Integer = false
  | equal_tya (Array ty) Void = false
  | equal_tya Void (Array ty) = false
  | equal_tya (Class literal) Void = false
  | equal_tya Void (Class literal) = false
  | equal_tya Nt Void = false
  | equal_tya Void Nt = false
  | equal_tya Integer Void = false
  | equal_tya Void Integer = false
  | equal_tya Boolean Void = false
  | equal_tya Void Boolean = false
  | equal_tya (Array tya) (Array ty) = equal_tya tya ty
  | equal_tya (Class literala) (Class literal) = ((literala : string) = literal)
  | equal_tya Nt Nt = true
  | equal_tya Integer Integer = true
  | equal_tya Boolean Boolean = true
  | equal_tya Void Void = true;

val equal_ty = {equal = equal_tya} : ty HOL.equal;

fun is_refT_i x =
  (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
    | Integer => Predicate.bot_pred | Nt => Predicate.single ()
    | Class _ => Predicate.single () | Array _ => Predicate.single ());

fun is_refT x1 = Predicate.holds (is_refT_i x1);

val throwable : string = "java/lang/Throwable";

fun ground_type (Array t) = ground_type t
  | ground_type Void = Void
  | ground_type Boolean = Boolean
  | ground_type Integer = Integer
  | ground_type Nt = Nt
  | ground_type (Class v) = Class v;

fun class_type_of (Class c) = SOME c
  | class_type_of (Array t) = SOME object
  | class_type_of Void = NONE
  | class_type_of Boolean = NONE
  | class_type_of Integer = NONE
  | class_type_of Nt = NONE;

end; (*struct Type*)

structure Transitive_Closure_Table : sig
  val rtrancl_tab_FioB_i_i_i :
    'a HOL.equal ->
      ('a -> 'a Predicate.pred) -> 'a list -> 'a -> 'a -> unit Predicate.pred
  val rtranclp_FioB_i_i :
    'a HOL.equal -> ('a -> 'a Predicate.pred) -> 'a -> 'a -> unit Predicate.pred
end = struct

fun rtrancl_tab_FioB_i_i_i A_ r xa xb xc =
  Predicate.sup_pred
    (if HOL.eq A_ xb xc then Predicate.single () else Predicate.bot_pred)
    (Predicate.bind (Predicate.if_pred (not (List.member A_ xa xb)))
      (fn () =>
        Predicate.bind (r xb)
          (fn xaa =>
            Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r (xb :: xa) xaa xc)
              (fn () => Predicate.single ()))));

fun rtranclp_FioB_i_i A_ r xa xb =
  Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r [] xa xb)
    (fn () => Predicate.single ());

end; (*struct Transitive_Closure_Table*)

structure TypeRel : sig
  datatype 'a fmod_ext = Fmod_ext of bool * 'a
  datatype 'a prog =
    Program of
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val class :
    'a prog ->
      string ->
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list)) option
  val eq_i_i : 'a HOL.equal -> 'a -> 'a -> unit Predicate.pred
  val classes :
    'a prog ->
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val is_class : 'a prog -> string -> bool
  val is_type : 'a prog -> Type.ty -> bool
  val subcls1_i_i_o : 'a prog -> string -> string Predicate.pred
  val widen_i_i_ia : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val widen : 'a prog -> Type.ty -> Type.ty -> bool
  val subcls_i_i_i : 'a prog -> string -> string -> unit Predicate.pred
  val subcls : 'a prog -> string -> string -> bool
  val is_class_i_i : 'a prog -> string -> unit Predicate.pred
  val widen_i_i_i : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val fields_i_i_o :
    'a prog ->
      string ->
        (((string * string) * (Type.ty * unit fmod_ext)) list) Predicate.pred
  val mapP_FioB_i_o :
    ('a -> 'b Predicate.pred) -> 'a list -> ('b list) Predicate.pred
  val methods_i_i_o :
    'a prog ->
      string ->
        (string -> ((Type.ty list * (Type.ty * 'a option)) * string) option)
          Predicate.pred
  val method_i_i_i_o_o_o_o :
    'a prog ->
      string ->
        string ->
          (Type.ty list * (Type.ty * ('a option * string))) Predicate.pred
  val sees_field_hoaux_PPiiii_Poooo :
    (string * string) * (Type.ty * unit fmod_ext) ->
      (string * (string * (Type.ty * unit fmod_ext))) Predicate.pred
  val sees_field_i_i_i_o_o_i :
    'a prog ->
      string -> string -> string -> (Type.ty * unit fmod_ext) Predicate.pred
end = struct

datatype 'a fmod_ext = Fmod_ext of bool * 'a;

datatype 'a prog =
  Program of
    (string *
      (string *
        ((string * (Type.ty * unit fmod_ext)) list *
          (string * (Type.ty list * (Type.ty * 'a option))) list))) list;

fun class (Program p) = Map.map_of Stringa.equal_literal p;

fun eq_i_i A_ xa xb =
  (if HOL.eq A_ xa xb then Predicate.single () else Predicate.bot_pred);

fun classes (Program p) = p;

fun is_class p c = not (Option.is_none (class p c));

fun is_type p Type.Void = true
  | is_type p Type.Boolean = true
  | is_type p Type.Integer = true
  | is_type p Type.Nt = true
  | is_type p (Type.Class c) = is_class p c
  | is_type p (Type.Array a) =
    (case Type.ground_type a of Type.Void => true | Type.Boolean => true
      | Type.Integer => true | Type.Nt => false | Type.Class aa => is_class p aa
      | Type.Array _ => true);

fun subcls1_i_i_o x xa =
  Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
    (fn () =>
      (case class x xa of NONE => Predicate.bot_pred
        | SOME (d, _) => Predicate.single d));

fun widen_i_i_ia x xa xb =
  Predicate.sup_pred
    (if Type.equal_tya xa xb then Predicate.single () else Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
      | Type.Nt =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.single ()
          | Type.Array _ => Predicate.single ())
      | Type.Class c =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class d =>
            Predicate.bind
              (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
                (subcls1_i_i_o x) c d)
              (fn () => Predicate.single ())
          | Type.Array _ => Predicate.bot_pred)
      | Type.Array t =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class c =>
            Predicate.bind (eq_i_i Stringa.equal_literal c Type.object)
              (fn () => Predicate.single ())
          | Type.Array ta =>
            Predicate.bind (widen_i_i_ia x t ta)
              (fn () => Predicate.single ())));

fun widen x1 x2 x3 = Predicate.holds (widen_i_i_ia x1 x2 x3);

fun subcls_i_i_i x xa xb =
  Predicate.bind
    (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
      (subcls1_i_i_o x) xa xb)
    (fn () => Predicate.single ());

fun subcls x1 x2 x3 = Predicate.holds (subcls_i_i_i x1 x2 x3);

fun is_class_i_i x xa =
  Predicate.bind (Predicate.if_pred (not (Option.is_none (class x xa))))
    (fn () => Predicate.single ());

fun widen_i_i_i p ta t =
  (if widen p ta t then Predicate.single () else Predicate.bot_pred);

fun fields_i_i_o x xa =
  Predicate.sup_pred
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (fs, _)) =>
            Predicate.bind (fields_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (List.map (fn (f, a) => ((f, xa), a)) fs @ xb)))))
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (fs, _)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (List.map (fn (f, a) => ((f, Type.object), a)) fs)
          else Predicate.bot_pred));

fun mapP_FioB_i_o h x =
  Predicate.sup_pred
    (case x of [] => Predicate.single [] | _ :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred
      | z :: xu =>
        Predicate.bind (mapP_FioB_i_o h xu)
          (fn xa =>
            Predicate.bind (h z) (fn xaa => Predicate.single (xaa :: xa))));

fun methods_i_i_o x xa =
  Predicate.sup_pred
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (_, ms)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (Option.map (fn m => (m, Type.object)) o
                   Map.map_of Stringa.equal_literal ms)
          else Predicate.bot_pred))
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (_, ms)) =>
            Predicate.bind (methods_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (Map.map_add xb
                    (Option.map (fn m => (m, xa)) o
                      Map.map_of Stringa.equal_literal ms))))));

fun method_i_i_i_o_o_o_o x xa xb =
  Predicate.bind (methods_i_i_o x xa)
    (fn xc =>
      (case xc xb of NONE => Predicate.bot_pred
        | SOME ((tv, (tc, p)), dc) => Predicate.single (tv, (tc, (p, dc)))));

fun sees_field_hoaux_PPiiii_Poooo ((x, xf), (xg, xh)) =
  Predicate.single (x, (xf, (xg, xh)));

fun sees_field_i_i_i_o_o_i x xa xb xc =
  Predicate.bind (fields_i_i_o x xa)
    (fn xd =>
      Predicate.bind (mapP_FioB_i_o sees_field_hoaux_PPiiii_Poooo xd)
        (fn xe =>
          (case Map.map_of Stringa.equal_literal xe xb
            of NONE => Predicate.bot_pred
            | SOME (dca, (tc, fp)) =>
              (if ((xc : string) = dca) then Predicate.single (tc, fp)
                else Predicate.bot_pred))));

end; (*struct TypeRel*)

structure BinOp : sig
  datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
    GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor
    | ShiftLeft | ShiftRightZeros | ShiftRightSigned
  val wT_binop_i_i_i_i_i :
    'a TypeRel.prog ->
      Type.ty -> bop -> Type.ty -> Type.ty -> unit Predicate.pred
  val wT_binop_i_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> bop -> Type.ty -> Type.ty Predicate.pred
end = struct

datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
  GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor |
  ShiftLeft | ShiftRightZeros | ShiftRightSigned;

fun wT_binop_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (case xb
      of Eq =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean =>
            Predicate.bind
              (Predicate.if_pred
                (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
              (fn () => Predicate.single ())
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | NotEq =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean =>
            Predicate.bind
              (Predicate.if_pred
                (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
              (fn () => Predicate.single ())
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
          | GreaterThan => Predicate.bot_pred
          | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
          | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
          | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft => Predicate.bot_pred
          | ShiftRightZeros => Predicate.bot_pred
          | ShiftRightSigned => Predicate.bot_pred)
      | Type.Integer =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | LessOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Add =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Subtract =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mult =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Div =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mod =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightZeros =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightSigned =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred))
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred);

fun wT_binop_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Eq =>
        Predicate.bind
          (Predicate.if_pred
            (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
          (fn () => Predicate.single Type.Boolean)
      | NotEq =>
        Predicate.bind
          (Predicate.if_pred
            (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
          (fn () => Predicate.single Type.Boolean)
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
          | GreaterThan => Predicate.bot_pred
          | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
          | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
          | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft => Predicate.bot_pred
          | ShiftRightZeros => Predicate.bot_pred
          | ShiftRightSigned => Predicate.bot_pred)
      | Type.Integer =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | LessOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Add =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Subtract =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mult =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Div =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mod =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightZeros =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightSigned =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred))
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred);

end; (*struct BinOp*)

structure Type_Length : sig
  type 'a len0
  val len_of : 'a len0 -> 'a HOL.itself -> IntInf.int
end = struct

type 'a len0 = {len_of : 'a HOL.itself -> IntInf.int};
val len_of = #len_of : 'a len0 -> 'a HOL.itself -> IntInf.int;

end; (*struct Type_Length*)

structure Word : sig
  datatype 'a word = Word of IntInf.int
end = struct

datatype 'a word = Word of IntInf.int;

end; (*struct Word*)

structure Numeral_Type : sig
  type 'a bit0
  type num1
end = struct

datatype 'a bit0 = EMPTY__

datatype num1 = EMPTY__

end; (*struct Numeral_Type*)

structure Value : sig
  datatype 'a vala = Unit | Null | Bool of bool |
    Intg of
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
    | Addr of 'a
end = struct

datatype 'a vala = Unit | Null | Bool of bool |
  Intg of
    Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
      Numeral_Type.bit0
      Numeral_Type.bit0
      Word.word
  | Addr of 'a;

end; (*struct Value*)

structure Expr : sig
  datatype ('a, 'b, 'c) exp = New of string |
    NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp
    | InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
    BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
    LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp
    | AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
    FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
    Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
    Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
    Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
    Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
    TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp
end = struct

datatype ('a, 'b, 'c) exp = New of string |
  NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp |
  InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
  BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
  LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
  FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
  Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
  Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
  Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
  Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
  TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp;

end; (*struct Expr*)

structure Heap : sig
  val typeof : 'a Value.vala -> Type.ty option
end = struct

fun typeof Value.Unit = SOME Type.Void
  | typeof Value.Null = SOME Type.Nt
  | typeof (Value.Bool b) = SOME Type.Boolean
  | typeof (Value.Intg i) = SOME Type.Integer
  | typeof (Value.Addr a) = NONE;

end; (*struct Heap*)

structure DefAss : sig
  val hyperUn : 'a Set.set option -> 'a Set.set option -> 'a Set.set option
  val hyperInt :
    'a HOL.equal -> 'a Set.set option -> 'a Set.set option -> 'a Set.set option
  val hyperDiff1 : 'a HOL.equal -> 'a Set.set option -> 'a -> 'a Set.set option
  val a : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> 'a Set.set option
  val a_s : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> 'a Set.set option
  val hyper_isin : 'a HOL.equal -> 'a -> 'a Set.set option -> bool
  val d : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> 'a Set.set option -> bool
  val d_s :
    'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> 'a Set.set option -> bool
end = struct

fun hyperUn a b =
  (case a of NONE => NONE
    | SOME aa => (case b of NONE => NONE | SOME ba => SOME (Set.union aa ba)));

fun hyperInt A_ a b =
  (case a of NONE => b
    | SOME aa =>
      (case b of NONE => SOME aa | SOME ba => SOME (Set.inf_set A_ aa ba)));

fun hyperDiff1 A_ aa a =
  (case aa of NONE => NONE
    | SOME aaa => SOME (Set.minus_set A_ aaa (Set.insert a Set.bot_set)));

fun a A_ (Expr.New c) = SOME Set.bot_set
  | a A_ (Expr.NewArray (t, e)) = a A_ e
  | a A_ (Expr.Cast (c, e)) = a A_ e
  | a A_ (Expr.InstanceOf (e, t)) = a A_ e
  | a A_ (Expr.Val v) = SOME Set.bot_set
  | a A_ (Expr.BinOp (e_1, bop, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Var v) = SOME Set.bot_set
  | a A_ (Expr.LAss (v, e)) = hyperUn (SOME (Set.insert v Set.bot_set)) (a A_ e)
  | a A_ (Expr.AAcc (aa, i)) = hyperUn (a A_ aa) (a A_ i)
  | a A_ (Expr.AAss (aa, i, e)) = hyperUn (hyperUn (a A_ aa) (a A_ i)) (a A_ e)
  | a A_ (Expr.ALen aa) = a A_ aa
  | a A_ (Expr.FAcc (e, f, d)) = a A_ e
  | a A_ (Expr.FAss (e_1, f, d, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Call (e, m, es)) = hyperUn (a A_ e) (a_s A_ es)
  | a A_ (Expr.Block (v, t, vo, e)) = hyperDiff1 A_ (a A_ e) v
  | a A_ (Expr.Synchronized (v, oa, e)) = hyperUn (a A_ oa) (a A_ e)
  | a A_ (Expr.InSynchronized (v, aa, e)) = a A_ e
  | a A_ (Expr.Seq (e_1, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Cond (e, e_1, e_2)) =
    hyperUn (a A_ e) (hyperInt A_ (a A_ e_1) (a A_ e_2))
  | a A_ (Expr.While (b, e)) = a A_ b
  | a A_ (Expr.Throw e) = NONE
  | a A_ (Expr.TryCatch (e_1, c, v, e_2)) =
    hyperInt A_ (a A_ e_1) (hyperDiff1 A_ (a A_ e_2) v)
and a_s A_ [] = SOME Set.bot_set
  | a_s A_ (e :: es) = hyperUn (a A_ e) (a_s A_ es);

fun hyper_isin A_ aa a =
  (case a of NONE => true | SOME ab => Set.member A_ aa ab);

fun d A_ (Expr.New c) a = true
  | d A_ (Expr.NewArray (t, e)) a = d A_ e a
  | d A_ (Expr.Cast (c, e)) a = d A_ e a
  | d A_ (Expr.InstanceOf (e, t)) a = d A_ e a
  | d A_ (Expr.Val v) a = true
  | d A_ (Expr.BinOp (e_1, bop, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Var v) a = hyper_isin A_ v a
  | d A_ (Expr.LAss (v, e)) a = d A_ e a
  | d A_ (Expr.AAcc (aaa, i)) aa =
    d A_ aaa aa andalso d A_ i (hyperUn aa (a A_ aaa))
  | d A_ (Expr.AAss (aaa, i, e)) aa =
    d A_ aaa aa andalso
      (d A_ i (hyperUn aa (a A_ aaa)) andalso
        d A_ e (hyperUn (hyperUn aa (a A_ aaa)) (a A_ i)))
  | d A_ (Expr.ALen aa) a = d A_ aa a
  | d A_ (Expr.FAcc (e, f, da)) a = d A_ e a
  | d A_ (Expr.FAss (e_1, f, da, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Call (e, m, es)) aa =
    d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e))
  | d A_ (Expr.Block (v, t, vo, e)) a =
    (if Option.is_none vo then d A_ e (hyperDiff1 A_ a v)
      else d A_ e (hyperUn a (SOME (Set.insert v Set.bot_set))))
  | d A_ (Expr.Synchronized (v, oa, e)) aa =
    d A_ oa aa andalso d A_ e (hyperUn aa (a A_ oa))
  | d A_ (Expr.InSynchronized (v, aa, e)) a = d A_ e a
  | d A_ (Expr.Seq (e_1, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Cond (e, e_1, e_2)) aa =
    d A_ e aa andalso
      (d A_ e_1 (hyperUn aa (a A_ e)) andalso d A_ e_2 (hyperUn aa (a A_ e)))
  | d A_ (Expr.While (e, c)) aa = d A_ e aa andalso d A_ c (hyperUn aa (a A_ e))
  | d A_ (Expr.Throw e) a = d A_ e a
  | d A_ (Expr.TryCatch (e_1, c, v, e_2)) a =
    d A_ e_1 a andalso d A_ e_2 (hyperUn a (SOME (Set.insert v Set.bot_set)))
and d_s A_ [] a = true
  | d_s A_ (e :: es) aa = d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e));

end; (*struct DefAss*)

structure While_Combinator : sig
  val while_option : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a option
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
end = struct

fun while_option b c s = (if b s then while_option b c (c s) else SOME s);

fun whilea b c s = Option.the (while_option b c s);

end; (*struct While_Combinator*)

structure SemiType : sig
  val super : 'a TypeRel.prog -> Type.ty -> Type.ty
  val widen1p_i_i_o : 'a TypeRel.prog -> Type.ty -> Type.ty Predicate.pred
  val sup : 'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Err.err
end = struct

fun super p (Type.Array Type.Integer) = Type.Class Type.object
  | super p (Type.Array Type.Boolean) = Type.Class Type.object
  | super p (Type.Array Type.Void) = Type.Class Type.object
  | super p (Type.Array (Type.Class c)) =
    (if ((c : string) = Type.object) then Type.Class Type.object
      else Type.Array (super p (Type.Class c)))
  | super p (Type.Array (Type.Array t)) = Type.Array (super p (Type.Array t))
  | super p (Type.Class c) =
    Type.Class (Product_Type.fst (Option.the (TypeRel.class p c)));

fun widen1p_i_i_o x xa =
  (case xa of Type.Void => Predicate.bot_pred
    | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
    | Type.Nt => Predicate.bot_pred
    | Type.Class c =>
      Predicate.bind (TypeRel.subcls1_i_i_o x c)
        (fn xb => Predicate.single (Type.Class xb))
    | Type.Array t =>
      Predicate.sup_pred
        (case t of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class xb =>
            (if ((xb : string) = Type.object)
              then Predicate.single (Type.Class Type.object)
              else Predicate.bot_pred)
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case t of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.single (Type.Class Type.object)
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case t of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single (Type.Class Type.object)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case t of Type.Void => Predicate.single (Type.Class Type.object)
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
              (Predicate.bind
                (Predicate.if_pred
                  (not (Type.equal_tya (Type.ground_type t) Type.Nt)))
                (fn () =>
                  Predicate.bind (widen1p_i_i_o x t)
                    (fn xb => Predicate.single (Type.Array xb))))))));

fun sup p t u =
  (if Type.is_refT t andalso Type.is_refT u
    then Err.Ok
           (if Type.equal_tya u Type.Nt then t
             else (if Type.equal_tya t Type.Nt then u
                    else While_Combinator.whilea
                           (fn z =>
                             not (Predicate.holds
                                   (Transitive_Closure_Table.rtrancl_tab_FioB_i_i_i
                                     Type.equal_ty (widen1p_i_i_o p) [] t z)))
                           (super p) u))
    else (if Type.equal_tya t u then Err.Ok t else Err.Err));

end; (*struct SemiType*)

structure Exceptions : sig
  val sys_xcpts_list : string list
  val wf_syscls : 'a TypeRel.prog -> bool
end = struct

val sys_xcpts_list : string list =
  ["java/lang/NullPointerException", "java/lang/ClassCastException",
    "java/lang/OutOfMemoryError", "java/lang/ArrayIndexOutOfBoundsException",
    "java/lang/ArrayStoreException", "java/lang/NegativeArraySizeException",
    "java/lang/ArithmeticException", "java/lang/IllegalMonitorStateException",
    "java/lang/IllegalThreadStateException", "java/lang/InterruptedException"];

fun wf_syscls p =
  List.list_all (TypeRel.is_class p)
    [Type.object, Type.throwable, Type.thread] andalso
    List.list_all (fn c => TypeRel.subcls p c Type.throwable) sys_xcpts_list;

end; (*struct Exceptions*)

structure ExternalCall : sig
  val external_WT_defs_i_i_i_i :
    string -> string -> Type.ty list -> Type.ty -> unit Predicate.pred
  val external_WT_defs : string -> string -> Type.ty list -> Type.ty -> bool
end = struct

fun external_WT_defs_i_i_i_i x xa xb xc =
  Predicate.sup_pred
    (case xb
      of [] =>
        (case xc
          of Type.Void =>
            (if ((xa : string) = "start()V") andalso
                  ((x : string) = Type.thread)
              then Predicate.single () else Predicate.bot_pred)
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb
        of [] =>
          (case xc
            of Type.Void =>
              (if ((xa : string) = "join()V") andalso
                    ((x : string) = Type.thread)
                then Predicate.single () else Predicate.bot_pred)
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xb
          of [] =>
            (case xc
              of Type.Void =>
                (if ((xa : string) = "interrupt()V") andalso
                      ((x : string) = Type.thread)
                  then Predicate.single () else Predicate.bot_pred)
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xb
            of [] =>
              (case xc of Type.Void => Predicate.bot_pred
                | Type.Boolean =>
                  (if ((xa : string) = "isInterrupted()Z") andalso
                        ((x : string) = Type.thread)
                    then Predicate.single () else Predicate.bot_pred)
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xb
              of [] =>
                (case xc
                  of Type.Void =>
                    (if ((xa : string) = "wait()V") andalso
                          ((x : string) = Type.object)
                      then Predicate.single () else Predicate.bot_pred)
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xb
                of [] =>
                  (case xc
                    of Type.Void =>
                      (if ((xa : string) = "notify()V") andalso
                            ((x : string) = Type.object)
                        then Predicate.single () else Predicate.bot_pred)
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xb
                  of [] =>
                    (case xc
                      of Type.Void =>
                        (if ((xa : string) = "notifyAll()V") andalso
                              ((x : string) = Type.object)
                          then Predicate.single () else Predicate.bot_pred)
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer => Predicate.bot_pred
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xb
                    of [] =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer => Predicate.bot_pred
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class xba =>
                          (if ((xba : string) = Type.object) andalso
                                (((xa : string) = "clone()Ljava/lang/Object;") andalso
                                  ((x : string) = Type.object))
                            then Predicate.single () else Predicate.bot_pred)
                        | Type.Array _ => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xb
                      of [] =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (if ((xa : string) = "hashCode()I") andalso
                                  ((x : string) = Type.object)
                              then Predicate.single () else Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xb of [] => Predicate.bot_pred
                        | Type.Void :: lista => Predicate.bot_pred
                        | Type.Boolean :: lista => Predicate.bot_pred
                        | [Type.Integer] =>
                          (case xc
                            of Type.Void =>
                              (if ((xa : string) = "~print(I)V") andalso
                                    ((x : string) = Type.object)
                                then Predicate.single ()
                                else Predicate.bot_pred)
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Type.Integer :: _ :: _ => Predicate.bot_pred
                        | Type.Nt :: lista => Predicate.bot_pred
                        | Type.Class _ :: lista => Predicate.bot_pred
                        | Type.Array _ :: lista => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xb
                          of [] =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.bot_pred
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class xba =>
                                (if ((xba : string) = Type.thread) andalso
                                      (((xa : string) = "~Thread.currentThread()Ljava/lang/Thread;") andalso
((x : string) = Type.object))
                                  then Predicate.single ()
                                  else Predicate.bot_pred)
                              | Type.Array _ => Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xb
                            of [] =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (if ((xa : string) = "~Thread.interrupted()Z") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (case xb
                            of [] =>
                              (case xc
                                of Type.Void =>
                                  (if ((xa : string) = "~Thread.yield()V") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred))))))))))));

fun external_WT_defs x1 x2 x3 x4 =
  Predicate.holds (external_WT_defs_i_i_i_i x1 x2 x3 x4);

end; (*struct ExternalCall*)

structure Predicate_Compile_Alternative_Defs : sig
  val list_all2_FiiB_i_i :
    ('a -> 'b -> unit Predicate.pred) ->
      'a list -> 'b list -> unit Predicate.pred
end = struct

fun list_all2_FiiB_i_i p xa xb =
  Predicate.sup_pred
    (case xa
      of [] =>
        (case xb of [] => Predicate.single () | _ :: _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (case xa of [] => Predicate.bot_pred
      | x :: xs =>
        (case xb of [] => Predicate.bot_pred
          | y :: ys =>
            Predicate.bind (list_all2_FiiB_i_i p xs ys)
              (fn () =>
                Predicate.bind (p x y) (fn () => Predicate.single ()))));

end; (*struct Predicate_Compile_Alternative_Defs*)

structure WellForm : sig
  val wf_fdecl :
    'a TypeRel.prog -> string * (Type.ty * unit TypeRel.fmod_ext) -> bool
  val wf_mdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_overriding_aux_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty ->
        Type.ty list -> Type.ty list -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty list -> string -> string -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_i_i_i :
    'a TypeRel.prog ->
      string ->
        string * (Type.ty list * (Type.ty * 'a option)) -> unit Predicate.pred
  val wf_overriding :
    'a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_cdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string *
          (string *
            ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
              (string * (Type.ty list * (Type.ty * 'a option))) list)) ->
          bool
  val acyclic_class_hierarchy : 'a TypeRel.prog -> bool
  val wf_prog :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog -> bool
end = struct

fun wf_fdecl p = (fn (_, (t, _)) => TypeRel.is_type p t);

fun wf_mdecl wf_md p c =
  (fn (m, (ts, (t, ma))) =>
    List.list_all (TypeRel.is_type p) ts andalso
      (TypeRel.is_type p t andalso
        (case ma of NONE => ExternalCall.external_WT_defs c m ts t
          | SOME mb => wf_md p c (m, (ts, (t, mb))))));

fun wf_overriding_aux_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (Predicate.bind
      (Predicate.not_pred
        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
          (TypeRel.widen_i_i_i xd) xc xb))
      (fn () => Predicate.single ()))
    (Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xd xa x))
      (fn () => Predicate.single ()));

fun wf_overriding_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o xd xc xb)
    (fn (ts, (t, (_, _))) =>
      Predicate.bind (wf_overriding_aux_aux_i_i_i_i_i t x xa ts xd)
        (fn () => Predicate.single ()));

fun wf_overriding_i_i_i x xa (xb, (xc, (xd, xe))) =
  Predicate.bind
    (Predicate.not_pred (wf_overriding_aux_i_i_i_i_i xd xc xb xa x))
    (fn () => Predicate.single ());

fun wf_overriding x1 x2 x3 = Predicate.holds (wf_overriding_i_i_i x1 x2 x3);

fun wf_cdecl wf_md p =
  (fn (c, (d, (fs, ms))) =>
    List.list_all (wf_fdecl p) fs andalso
      (Aux.distinct_fst Stringa.equal_literal fs andalso
        (List.list_all (wf_mdecl wf_md p c) ms andalso
          (Aux.distinct_fst Stringa.equal_literal ms andalso
            ((if not ((c : string) = Type.object)
               then TypeRel.is_class p d andalso
                      List.list_all (wf_overriding p d) ms
               else true) andalso
              (if ((c : string) = Type.thread)
                then List.member
                       (Product_Type.equal_prod Stringa.equal_literal
                         (Product_Type.equal_prod
                           (List.equal_list Type.equal_ty) Type.equal_ty))
                       (List.map (fn (m, (ts, (t, _))) => (m, (ts, t))) ms)
                       ("run()V", ([], Type.Void))
                else true))))));

fun acyclic_class_hierarchy p =
  List.list_all
    (fn (c, (d, (_, _))) =>
      (if not ((c : string) = Type.object) then not (TypeRel.subcls p d c)
        else true))
    (TypeRel.classes p);

fun wf_prog wf_md p =
  acyclic_class_hierarchy p andalso
    (Exceptions.wf_syscls p andalso
      (Aux.distinct_fst Stringa.equal_literal (TypeRel.classes p) andalso
        List.list_all (wf_cdecl wf_md p) (TypeRel.classes p)));

end; (*struct WellForm*)

structure WellType : sig
  val wT_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
  val wTs_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp list -> (Type.ty list) Predicate.pred
  val wT_Fii_oB_i_i_i_i :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val is_lub_sup_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Predicate.pred
  val wT_code_i_i_i_o :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
end = struct

fun wT_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb
    of Expr.New c1 =>
      Predicate.bind (TypeRel.is_class_i_i x c1)
        (fn () => Predicate.single (Type.Class c1))
    | Expr.NewArray (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
            (fn () => Predicate.single (Type.Array c1)))
    | Expr.Cast (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c1))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x c1 xaa orelse TypeRel.widen x xaa c1))
                (fn () => Predicate.single c1)))
    | Expr.InstanceOf (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind (Type.is_refT_i c2)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn xaa =>
                  Predicate.bind
                    (Predicate.if_pred
                      (TypeRel.widen x c2 xaa orelse TypeRel.widen x xaa c2))
                    (fn () => Predicate.single Type.Boolean))))
    | Expr.Val c1 =>
      (case Heap.typeof c1 of NONE => Predicate.bot_pred
        | SOME a => Predicate.single a)
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (BinOp.wT_binop_i_i_i_i_o x xba c2))
    | Expr.Var c1 =>
      (case xa c1 of NONE => Predicate.bot_pred | SOME a => Predicate.single a)
    | Expr.LAss (c1, c2) =>
      Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              (case xa c1 of NONE => Predicate.bot_pred
                | SOME t =>
                  Predicate.bind (TypeRel.widen_i_i_i x xba t)
                    (fn () => Predicate.single Type.Void))))
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array aa => Predicate.single aa)))
    | Expr.AAss (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array t =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                    (fn xaa =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.ALen c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.single Type.Integer))
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          (case Type.class_type_of xaa of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                (fn (t, _) => Predicate.single t)))
    | Expr.FAss (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
            (fn xaa =>
              (case Type.class_type_of xba of NONE => Predicate.bot_pred
                | SOME c =>
                  Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                    (fn (t, _) =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          (case Type.class_type_of xba of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x c c2)
                    (fn (ts, (t, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single t)))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind
            (Predicate.if_pred
              (case c3 of NONE => true
                | SOME v =>
                  (case Heap.typeof v of NONE => false
                    | SOME t => TypeRel.widen x t c2)))
            (fn () =>
              wT_Fii_oB_i_i_i_o is_lub x
                (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (Type.is_refT_i xba)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (not (Type.equal_tya xba Type.Nt)))
                (fn () => wT_Fii_oB_i_i_i_o is_lub x xa c3)))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn _ => wT_Fii_oB_i_i_i_o is_lub x xa c2)
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3) (is_lub xba)))
    | Expr.While (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn _ => Predicate.single Type.Void))
    | Expr.Throw c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class c =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                  Type.throwable)
                (fn () => Predicate.single Type.Void)
            | Type.Array _ => Predicate.bot_pred))
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind
        (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
          (TypeRel.subcls1_i_i_o x) c2 Type.throwable)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind
                (wT_Fii_oB_i_i_i_i is_lub x
                  (Fun.fun_upd Stringa.equal_literal xa c3
                    (SOME (Type.Class c2)))
                  c4 xba)
                (fn () => Predicate.single xba))))
and wTs_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb of [] => Predicate.single []
    | c1 :: c2 =>
      Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xc => Predicate.single (xc :: xba))))
and wT_Fii_oB_i_i_i_i is_lub x xa xb xc =
  (case xb
    of Expr.New _ =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class c1 =>
          Predicate.bind (TypeRel.is_class_i_i x c1)
            (fn () => Predicate.single ())
        | Type.Array _ => Predicate.bot_pred)
    | Expr.NewArray (_, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array c1 =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
                (fn () => Predicate.single ())))
    | Expr.Cast (_, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x xc xaa orelse TypeRel.widen x xaa xc))
                (fn () => Predicate.single ())))
    | Expr.InstanceOf (c1, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind (Type.is_refT_i c2)
                (fn () =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                    (fn xaa =>
                      Predicate.bind
                        (Predicate.if_pred
                          (TypeRel.widen x c2 xaa orelse
                            TypeRel.widen x xaa c2))
                        (fn () => Predicate.single ()))))
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Val c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (Heap.typeof c1)
          (SOME xc))
        (fn () => Predicate.single ())
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (fn xaa =>
              Predicate.bind (BinOp.wT_binop_i_i_i_i_i x xba c2 xaa xc)
                (fn () => Predicate.single ())))
    | Expr.Var c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (xa c1) (SOME xc))
        (fn () => Predicate.single ())
    | Expr.LAss (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn xba =>
                  (case xa c1 of NONE => Predicate.bot_pred
                    | SOME t =>
                      Predicate.bind (TypeRel.widen_i_i_i x xba t)
                        (fn () => Predicate.single ()))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 (Type.Array xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () => Predicate.single ()))
    | Expr.AAss (c1, c2, c3) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn a =>
                  (case a of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array t =>
                      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                        (fn xaa =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.ALen c1 =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.single ()))
        | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          (case Type.class_type_of xaa of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                (fn (_, _) => Predicate.single ())))
    | Expr.FAss (c1, c2, c3, c4) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
                (fn xaa =>
                  (case Type.class_type_of xba of NONE => Predicate.bot_pred
                    | SOME c =>
                      Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                        (fn (t, _) =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          (case Type.class_type_of xba of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x c c2)
                    (fn (ts, (_, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single ())))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind
        (wT_Fii_oB_i_i_i_i is_lub x
          (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4 xc)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind
                (Predicate.if_pred
                  (case c3 of NONE => true
                    | SOME v =>
                      (case Heap.typeof v of NONE => false
                        | SOME t => TypeRel.widen x t c2)))
                (fn () => Predicate.single ())))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c3 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xd =>
              Predicate.bind (Type.is_refT_i xd)
                (fn () =>
                  Predicate.bind
                    (Predicate.if_pred (not (Type.equal_tya xd Type.Nt)))
                    (fn () => Predicate.single ()))))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn _ => Predicate.single ()))
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xd =>
                  Predicate.bind (is_lub xba xd)
                    (fn _ => Predicate.single ()))))
    | Expr.While (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn _ => Predicate.single ()))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Throw c1 =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class c =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                      Type.throwable)
                    (fn () => Predicate.single ())
                | Type.Array _ => Predicate.bot_pred))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 xc)
        (fn () =>
          Predicate.bind
            (wT_Fii_oB_i_i_i_i is_lub x
              (Fun.fun_upd Stringa.equal_literal xa c3 (SOME (Type.Class c2)))
              c4 xc)
            (fn () =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c2
                  Type.throwable)
                (fn () => Predicate.single ()))));

fun is_lub_sup_i_i_i_o x xa xb =
  (case SemiType.sup x xa xb of Err.Err => Predicate.bot_pred
    | Err.Ok a => Predicate.single a);

fun wT_code_i_i_i_o x xa xb = wT_Fii_oB_i_i_i_o (is_lub_sup_i_i_i_o x) x xa xb;

end; (*struct WellType*)

structure JWellForm : sig
  val typeable_with_i_i_i_i :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val typeable_with :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> bool
  val wf_J_mdecl :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      string ->
        string *
          (Type.ty list *
            (Type.ty * (string list * (string, unit, 'a) Expr.exp))) ->
          bool
  val wf_J_prog :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog -> bool
end = struct

fun typeable_with_i_i_i_i x xa xb xc =
  Predicate.bind (WellType.wT_code_i_i_i_o x xa xb)
    (fn xaa =>
      Predicate.bind (TypeRel.widen_i_i_i x xaa xc)
        (fn () => Predicate.single ()));

fun typeable_with x1 x2 x3 x4 =
  Predicate.holds (typeable_with_i_i_i_i x1 x2 x3 x4);

fun wf_J_mdecl p c =
  (fn (_, (ts, (t, (pns, body)))) =>
    (((List.size_list ts) : IntInf.int) = (List.size_list pns)) andalso
      (List.distinct Stringa.equal_literal pns andalso
        (not (List.member Stringa.equal_literal pns "this") andalso
          (typeable_with p
             (Map.map_upds Stringa.equal_literal
               (Fun.fun_upd Stringa.equal_literal (fn _ => NONE) "this"
                 (SOME (Type.Class c)))
               pns ts)
             body t andalso
            DefAss.d Stringa.equal_literal body
              (SOME (Set.union (Set.insert "this" Set.bot_set)
                      (Set.Set pns)))))));

fun wf_J_prog x = WellForm.wf_prog wf_J_mdecl x;

end; (*struct JWellForm*)

structure Code_Generation : sig
  val j_Program :
    (string *
      (string *
        ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
          (string *
            (Type.ty list *
              (Type.ty *
                (string list *
                  (string, unit, IntInf.int) Expr.exp) option))) list))) list ->
      (string list * (string, unit, IntInf.int) Expr.exp) TypeRel.prog
end = struct

fun j_Program x = TypeRel.Program x;

end; (*struct Code_Generation*)
